@@ -50,14 +50,44 @@ print.snowprofile <- function(x, pretty = TRUE, nLayers = NA, ...) {
   if (pretty) {
     vars <- x
     vars$layers <- NULL
+    if ("temperatureProfile" %in% names(vars)) vars$temperatureProfile <- NULL
+    if ("tests" %in% names(vars)) vars$tests <- NULL
+    if ("instabilitySigns" %in% names(vars)) vars$instabilitySigns <- NULL
+    if ("backtrackingTable" %in% names(vars)) vars$backtrackingTable <- "available, but omitted here"
+    if ("sim2group" %in% names(vars)) vars$sim2group <- "available, but omitted here"
     MetaData <- sapply(vars, function(i) paste(i, collapse = " "))
     print(as.data.frame(MetaData))
     if (is.na(nLayers)) {
+      cat("layers\n")
       print(x$layers)
+      if ("temperatureProfile" %in% names(x)) {
+        cat("temperatureProfile\n")
+        print(x$temperatureProfile)
+      }
+      if ("tests" %in% names(x)) {
+        cat("tests\n")
+        print(x$tests)
+      }
+      if ("instabilitySigns" %in% names(x)) {
+        cat("instabilitySigns\n")
+        print(x$instabilitySigns)
+      }
     } else {
       print(head(x$layers, nLayers))
       ndiff <- nrow(x$layers) - nLayers
       if (ndiff > 0) print(paste0("  -- ", ndiff, " layers omitted --  "))
+      if ("temperatureProfile" %in% names(x)) {
+        cat("temperatureProfile\n")
+        print(head(x$temperatureProfile, nLayers))
+      }
+      if ("tests" %in% names(x)) {
+        cat("tests\n")
+        print(x$tests)
+      }
+      if ("instabilitySigns" %in% names(x)) {
+        cat("instabilitySigns\n")
+        print(x$instabilitySigns)
+      }
     }
   } else {
     print.default(x)
@@ -69,19 +99,7 @@ print.snowprofile <- function(x, pretty = TRUE, nLayers = NA, ...) {
 #'
 #' Low-cost, efficient constructor function to be used by users who know what they're doing. If that's not you,
 #' use the high-level constructor [snowprofile].
-#' @param station character string
-#' @param station_id character string
-#' @param datetime date and time as class POSIXct in most meaningful timezone (timezone can be converted very easily:
-#' e.g. `print(profile$datetime, tz = 'EST')` defaults to `'1999-12-31 UTC'`
-#' @param latlon 2-element vector latitude (first), longitude (second)
-#' @param elev profile elevation (m)
-#' @param angle slope angle (degree)
-#' @param aspect slope aspect (degree)
-#' @param hs total snow height (cm); if not provided, the field will be derived from the profile layers.
-#' @param type character string, must be either 'manual', 'vstation' or 'aggregate'
-#' @param band character string describing elevation band as ALP, TL, BTL (alpine, treeline, below treeline)
-#' @param zone character string describing the zone or region of the profile location (e.g., BURNABY_MTN)
-#' @param layers [snowprofileLayers] object
+#' @inheritParams snowprofile
 #'
 #' @return snowprofile object
 #'
@@ -93,10 +111,17 @@ new_snowprofile <- function(station = character(),
                             angle = double(),
                             aspect = double(),
                             hs = double(),
+                            maxObservedDepth = double(),
                             type = character(),
                             band = character(),
                             zone = character(),
-                            layers = snowprofileLayers()) {
+                            comment = character(),
+                            hn24 = double(),
+                            hn72 = double(),
+                            ski_pen = double(),
+                            layers = snowprofileLayers(),
+                            tests = snowprofileTests(),
+                            instabilitySigns = snowprofileInstabilitySigns()) {
 
   ## check input types: omitted to save resources!
 
@@ -112,10 +137,17 @@ new_snowprofile <- function(station = character(),
                  angle = angle,
                  aspect = aspect,
                  hs = hs,
+                 maxObservedDepth = maxObservedDepth,
                  type = type,
                  band = band,
                  zone = zone,
-                 layers = layers)
+                 comment = comment,
+                 hn24 = hn24,
+                 hn72 = hn72,
+                 ski_pen = ski_pen,
+                 layers = layers,
+                 tests = tests,
+                 instabilitySigns = instabilitySigns)
 
   class(object) <- "snowprofile"
 
@@ -137,10 +169,18 @@ new_snowprofile <- function(station = character(),
 #' @param angle slope angle (degree)
 #' @param aspect slope aspect (degree)
 #' @param hs total snow height (cm); if not provided, the field will be derived from the profile layers.
+#' @param maxObservedDepth equivalent to `hs` for full profiles that go down to the ground. for test profiles that only
+#' observe the upper part of the snowpack this value refers to the maximum depth of the profile observation.
 #' @param type character string, must be either 'manual', 'modeled', 'vstation', 'aggregate', or 'whiteboard'
 #' @param band character string describing elevation band as ALP, TL, BTL (alpine, treeline, below treeline)
 #' @param zone character string describing the zone or region of the profile location (e.g., BURNABY_MTN)
+#' @param comment character string with any text comments
+#' @param hn24 height of new snow within 24 h
+#' @param hn72 height of new snow within 72 h
+#' @param ski_pen skier penetration depth (cm)
 #' @param layers [snowprofileLayers] object
+#' @param tests [snowprofileTests] object
+#' @param instabilitySigns [snowprofileInstabilitySigns] object
 #' @param validate Validate the object with [validate_snowprofile]?
 #' @param dropNAs Do you want to drop non-mandatory `snowprofile` and `snowprofileLayers` fields that are `NA` only?
 #'
@@ -148,7 +188,7 @@ new_snowprofile <- function(station = character(),
 #'
 #' @author shorton, fherla
 #'
-#' @seealso [summary.snowprofile], [plot.snowprofile], [snowprofileLayers], [SPpairs]
+#' @seealso [summary.snowprofile], [plot.snowprofile], [snowprofileLayers], [snowprofileTests], [snowprofileInstabilitySigns], [SPpairs]
 #'
 #' @examples
 #'
@@ -169,16 +209,23 @@ new_snowprofile <- function(station = character(),
 #'
 snowprofile <- function(station = as.character(NA),
                         station_id = as.character(NA),
-                        datetime = as.POSIXct(NA, tz = "UTC"),
+                        datetime = as.POSIXct(NA),
                         latlon = as.double(c(NA, NA)),
                         elev = as.double(NA),
                         angle = as.double(NA),
                         aspect = as.double(NA),
                         hs = as.double(NA),
+                        maxObservedDepth = as.double(NA),
                         type = "manual",
                         band = as.character(NA),
                         zone = as.character(NA),
-                        layers = snowprofileLayers(formatTarget = "FALSE", dropNAs = FALSE),
+                        comment = as.character(NA),
+                        hn24 = as.double(NA),
+                        hn72 = as.double(NA),
+                        ski_pen = as.double(NA),
+                        layers = snowprofileLayers(dropNAs = FALSE, validate = FALSE),
+                        tests = snowprofileTests(dropNAs = FALSE),
+                        instabilitySigns = snowprofileInstabilitySigns(dropNAs = FALSE),
                         validate = TRUE,
                         dropNAs = TRUE) {
 
@@ -189,20 +236,49 @@ snowprofile <- function(station = as.character(NA),
   #                               proj4string = sp::CRS('+proj=longlat +datum=WGS84'))
   # }
 
-  ## format hs:
+  ## Check and format hs and maxObservedDepth:
+  cols_layers <- colnames(layers)
+  vloc_max <- suppressWarnings( max(c(ifelse("height" %in% cols_layers, max(layers$height), NA),
+                                      ifelse(all(c("depth", "thickness") %in% cols_layers), layers$depth[1] + layers$thickness[1], NA)),
+                                    na.rm = TRUE) )
+  vloc_max <- ifelse(is.infinite(vloc_max), as.double(NA), vloc_max)
+  ## query for basal layer thickness NA --> hs was unknown when snowprofileLayers object was created!
+  if (is.na(layers$thickness[1])) {
+    maxObservedDepth_from_layers <- layers$depth[1]
+    hs_from_layers <- as.double(NA)
+  } else if (hasUnobservedBasalLayer(layers)) {
+    ## in this case, hs was known, but basal layer(s) not observed down to ground!
+    maxObservedDepth_from_layers <- layers$depth[1]
+    hs_from_layers <- vloc_max
+  } else {
+    maxObservedDepth_from_layers <- vloc_max
+    hs_from_layers <- vloc_max
+  }
+
   if (is.na(hs)) {
-    cols_layers <- colnames(layers)
-    if ("height" %in% cols_layers) {
-      hs <- max(layers$height)
-    } else if (all(c("depth", "thickness") %in% cols_layers)) {
-      argmax <- which.max(layers$depth)
-      hs <- layers$depth[argmax] + layers$thickness[argmax]
+    hs <- hs_from_layers
+  } else {
+    if (!isTRUE(all.equal(hs, hs_from_layers))) {
+      hs_new <- max(c(hs, hs_from_layers), na.rm = TRUE)
+      warning(paste0("hs (", hs, " cm) is different than hs deducted from your provided layers (", hs_from_layers, " cm)\n",
+                     "--> Setting hs to ", hs_new, " cm."))
+      hs <- hs_new
+    }
+  }
+  if (is.na(maxObservedDepth)) {
+    maxObservedDepth <- maxObservedDepth_from_layers
+  } else {
+    if (!isTRUE(all.equal(maxObservedDepth, maxObservedDepth_from_layers))) {
+      maxObservedDepth_new <- max(c(maxObservedDepth, maxObservedDepth_from_layers), na.rm = TRUE)
+      warning(paste0("maxObservedDepth (", maxObservedDepth, " cm) is different than maxObservedDepth deducted from your provided layers (", maxObservedDepth_from_layers, " cm)\n",
+                     "--> Setting maxObservedDepth to ", maxObservedDepth_new, " cm."))
+      maxObservedDepth <- maxObservedDepth_new
     }
   }
 
-  ## type conversion:
-  sapply(c(type), function(x) x <- as.character(x))
-  sapply(c(latlon, elev, angle, aspect, hs), function(x) x <- as.double(x))
+  ## type conversion:  NOTE: the two lines are uncommented b/c the code doesn't actually convert the type. Left to be done!
+  # sapply(c(type), function(x) x <- as.character(x))
+  # sapply(c(latlon, elev, angle, aspect, hs, maxObservedDepth), function(x) x <- as.double(x))
 
   ## create snowprofile object
   object <- new_snowprofile(station = station,
@@ -213,21 +289,30 @@ snowprofile <- function(station = as.character(NA),
                             angle = angle,
                             aspect = aspect,
                             hs = hs,
+                            maxObservedDepth = maxObservedDepth,
                             type = type,
                             band = band,
                             zone = zone,
-                            layers = layers)
+                            comment = comment,
+                            hn24 = hn24,
+                            hn72 = hn72,
+                            ski_pen = ski_pen,
+                            layers = layers,
+                            tests = tests,
+                            instabilitySigns = instabilitySigns)
 
   ## add date from datetime
-  object$date <- as.Date(object$datetime)
+  object$date <- as.Date(as.character(object$datetime))  # double conversion ensures that timezone won't change from POSIXct to Date format!
 
   ## define mandatory fields:
+  ## IMPORTANT: needs to be synced manually with mandatory_fields defined in snowprofile validator!!
   mandatory_fields <- c("station", "station_id", "datetime", "date", "latlon",
-                        "elev", "angle", "aspect", "hs", "type", "layers")
+                        "elev", "angle", "aspect", "hs", "maxObservedDepth", "type", "layers")
 
   ## drop NAs and run checks:
   if (dropNAs) {
-    object <- c(object[mandatory_fields], object[which(!is.na(object) & !names(object) %in% mandatory_fields)])
+    toKeep <- unique(c(mandatory_fields, names(which(!(sapply(object, function(x) (all(is.na(x)) || isTRUE(all(dim(x)[[1]] == 0))) ))))))
+    object <- object[toKeep]
     class(object) <- "snowprofile"
   }
   if (validate) validate_snowprofile(object)
@@ -279,6 +364,7 @@ snowprofile <- function(station = as.character(NA),
 #' ## so that you get a comprehensive list of errors of all profiles:
 #' if (!this_throws_error) {
 #' errorlist <- lapply(SPmalformatted, validate_snowprofile, silent = TRUE)
+#' errorlist[sapply(errorlist, function(item) !is.null(item))]  # print profiles that caused errors
 #' }
 #'
 #' @seealso [reformat_snowprofile]
@@ -292,10 +378,15 @@ validate_snowprofile <- function(object, silent = FALSE) {
   ## initialize error string:
   err <- NULL
   objNames <- names(object)
-  knownNames <- names(snowprofile(validate = FALSE, dropNAs = FALSE))
+  objNames <- objNames[order(objNames)]  # sort alphabetically
+  knownNames <- c(names(snowprofile(validate = FALSE, dropNAs = FALSE)), "psum", "psum24")
+  knownNames <- knownNames[order(knownNames)]  # sort alphabetically
 
-  ## mandatory fields:
-  mandatory_fields <- names(snowprofile(validate = FALSE, dropNAs = TRUE))
+
+  ## mandatory fields:  IMPORTANT: needs to be synced manually with mandatory_fields defined in snowprofile constructor!!
+  mandatory_fields <- mandatory_fields <- c("station", "station_id", "datetime", "date", "latlon",
+                                            "elev", "angle", "aspect", "hs", "maxObservedDepth", "type", "layers")
+  mandatory_fields <- mandatory_fields[order(mandatory_fields)]  # sort alphabetically
   field_check <- mandatory_fields %in% objNames
   if (!all(field_check))
     err <- paste(err, paste("Missing mandatory snowprofile fields:",
@@ -321,7 +412,7 @@ validate_snowprofile <- function(object, silent = FALSE) {
   try(if (!(object$type %in% c("manual", "modeled", "aggregate", "vstation", "whiteboard")))
     err <- paste(err, "type must be either manual, modeled, vstation, aggregate, or whiteboard", sep = "\n "))
 
-  for (x in c("elev", "angle", "aspect", "hs")) {
+  for (x in c("elev", "angle", "aspect", "hs", "maxObservedDepth")) {
     try(if (!is.numeric(object[[x]]))
       err <- paste(err, paste(x, "needs to be numeric"), sep = "\n "))
   }
@@ -390,7 +481,7 @@ reformat_snowprofile <- function(profile, currentFields = NULL, targetFields = N
   if (!is.snowprofile(profile)) stop("Not a class snowprofile object")
   if (length(currentFields) != length(targetFields)) stop("currentFields and targetFields need to be of same length!")
 
-  emptyPro <- snowprofile()
+  emptyPro <- snowprofile(validate = FALSE, dropNAs = FALSE)
   metaNames <- names(emptyPro)[names(emptyPro) != "layers"]
   metaTypes <- sapply(emptyPro[metaNames], class)
   layerNames <- names(emptyPro$layers)
@@ -413,15 +504,19 @@ reformat_snowprofile <- function(profile, currentFields = NULL, targetFields = N
   ## check which field types yield errors:
   msg <- tryCatch({validate_snowprofile(profile)},
                    error = function(e) e$message)
+
   if (length(msg) > 0) {
     msgSplit <- strsplit(msg, split = " ")[[1]]
     metaConvert <- msgSplit[sapply(msgSplit, `%in%`, metaNames)]
     layerConvert <- msgSplit[sapply(msgSplit, `%in%`, layerNames)]
 
+    if ("instabilitySigns" %in% metaConvert) stop("$instabilitySigns contain errors, fix manually!")
+    if ("tests" %in% metaConvert) stop("$tests contain errors, fix manually!")
+
     ## convert those field types according to types in snowprofile() constructor function
     if (length(metaConvert) > 0) {
       if (length(profile$latlon) != 2) stop("latlon needs to be a vector of length 2!")
-      if ("date" %in% metaConvert) try(profile$date <- as.Date(profile$datetime))
+      if ("date" %in% metaConvert) try(profile$date <- as.Date(as.character(profile$datetime)))
       for (mconv in metaConvert) {
         profile[mconv] <- do.call(paste0("as.", metaTypes[mconv][[1]][[1]]), unname(profile[mconv]))
       }
@@ -429,7 +524,19 @@ reformat_snowprofile <- function(profile, currentFields = NULL, targetFields = N
 
     if (length(layerConvert) > 0) {
       for (lconv in layerConvert) {
+        if (lconv %in% "gtype") {
+          ## if gtype doesn't exist --> create it with NAs
+          if (!"gtype" %in% names(profile$layers)) profile$layers$gtype <- as.factor(rep(as.character(NA), times = nrow(profile$layers)))
+        }
         profile$layers[lconv] <- do.call(paste0("as.", layerTypes[lconv][[1]][[1]]), unname(profile$layers[lconv]))
+        if (lconv %in% c("gtype","gtype_sec")){
+          unknown_gtypes <- which(!profile$layers[[lconv]] %in% c(grainDict$gtype, NA_character_))
+          if (length(unknown_gtypes > 0)) warning("There are be unknown grain types in your profile...simplifying them to their parent classes!")
+          converted_gtypes <- substr(as.character(profile$layers[[lconv]][unknown_gtypes]), 1, 2)
+          profile$layers[lconv] <- as.character(profile$layers[[lconv]])
+          profile$layers[[lconv]][unknown_gtypes] <- converted_gtypes
+          profile$layers[[lconv]] <- as.factor(profile$layers[[lconv]])
+        }
       }
     }
   }
